import aioamqp
import asyncio
import pickle
import uuid


from . import logger
from .utils import AsyncMixin


class Queue(AsyncMixin):

    def __init__(self, coro, loop=None, exchange='donald', queue='donald', **params):
        self.params = params

        self._coro = coro
        self._loop = loop or asyncio.get_event_loop()
        self._queue = queue
        self._transport = None
        self._protocol = None
        self._channel = None
        self._connected = False

    def is_connected(self):
        return self._connected

    @asyncio.coroutine
    def start(self, listen=True):
        """Connect to message queue."""
        logger.warn('Connect to queue.')
        try:
            self._transport, self._protocol = yield from aioamqp.connect(**self.params)
            self._channel = yield from self._protocol.channel()

            yield from self._channel.queue_declare(queue_name=self._queue, durable=True)
            yield from self._channel.basic_qos(
                prefetch_count=1, prefetch_size=0, connection_global=False)
            self._connected = True
            if listen:
                yield from self.listen()

        except aioamqp.AmqpClosedConnection as exc:
            logger.exception(exc)
            logger.error('Connection is closed.')

    def listen(self):
        """Run tasks from self.queue.

        :returns: A coroutine
        """
        return self._channel.basic_consume(self.callback, queue_name=self._queue)

    @asyncio.coroutine
    def stop(self):
        if not self.is_connected() or self.is_closed():
            return False

        logger.warn('Disconnect from queue.')

        yield from self._protocol.close()
        self._transport.close()
        self._connected = False

    @asyncio.coroutine
    def submit(self, func, *args, **kwargs):
        logger.info('Submit task to queue.')
        yield from self._channel.basic_publish(
            payload=pickle.dumps((func, args, kwargs)),
            exchange_name='', routing_key=self._queue,
            properties=dict(delivery_mode=2, message_id=str(uuid.uuid4()))
        )

    @asyncio.coroutine
    def callback(self, channel, body, envelope, properties):
        func, args, kwargs = pickle.loads(body)
        result = yield from self._coro.submit(func, *args, **kwargs)
        print('Get result %r %r' % (result, properties.message_id))
        yield from channel.basic_client_ack(delivery_tag=envelope.delivery_tag)



import logging


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


from .core import Donald # noqa


__version__ = "0.0.0"
__project__ = "Donald"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "BSD"



from concurrent.futures import Future


class AsyncMixin:

    def is_running(self):
        return self._loop and self._loop.is_running()

    def is_closed(self):
        return not self._loop or self._loop.is_closed()

    @property
    def loop(self):
        return self._loop


class AttrDict(dict):

    def __init__(self, *args, **kwargs):
        self.__dict__ = self
        super(AttrDict, self).__init__(*args, **kwargs)


class CallableFuture(Future):

    def __init__(self, func, *args, **kwargs):
        super(CallableFuture, self).__init__()
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def __call__(self):
        with self._condition:
            try:
                self.set_result(self._func(*self._args, **self._kwargs))
            except Exception as exc:
                self.set_exception(exc)



"""Main Donald class."""

import asyncio
import atexit
import datetime

from . import logger
from .queue import Queue
from .utils import AsyncMixin, AttrDict
from .worker import AsyncThreadWorker, call_with_loop


class Donald(AsyncMixin):

    """I'am on Donald."""

    defaults = dict(

        # Run tasks imediatelly
        always_eager=False,

        # Number of workers
        num_threads=4,

        # logging level
        loglevel='INFO',

    )

    def __init__(self, loop=None, **params):
        """Initialize donald parameters."""
        self.params = AttrDict(self.defaults)
        self.params.update(params)

        logger.setLevel(self.params['loglevel'].upper())

        # Base loop and queue
        self._loop = loop or asyncio.get_event_loop()
        num_threads = 1 if self.params.num_threads < 1 else self.params.num_threads
        self._threads = tuple(AsyncThreadWorker() for _ in range(num_threads))
        self._tqueue = asyncio.Queue(maxsize=len(self._threads), loop=self._loop)
        self._schedules = []
        self._closing = False

        self.queue = Queue(self, loop=self._loop)
        self.params.always_eager = self.params.always_eager or not len(self._threads)

    def start(self):
        """Start workers.

        :returns: A coroutine
        """
        logger.warn('Start Donald')

        atexit.register(self.stop)

        futures = []
        for t in self._threads:
            f = asyncio.wrap_future(t.start(), loop=self._loop)
            f.add_done_callback(lambda f: self._tqueue.put_nowait(t))
            futures.append(f)

        self.closing = False
        return asyncio.wait(futures)

    def stop(self):
        """Stop workers. Disconnect from queue. Cancel schedules.

        The method could be called syncroniously too.

        :returns: A future
        """
        if self.is_closed() or self.closing:
            return False

        logger.warn('Stop Donald.')

        for task in self._schedules:
            task.cancel()

        # Stop queue
        tasks = [asyncio.ensure_future(self.queue.stop())]

        # Stop workers
        tasks += [
            asyncio.ensure_future(asyncio.wrap_future(t.stop(), loop=self.loop))
            for t in self._threads]

        self.closing = True
        return asyncio.wait([self.queue.stop()])

    @asyncio.coroutine
    def submit(self, func, *args, **kwargs):
        """Submit given function/coroutine."""
        if self.params.always_eager:
            future = call_with_loop(self._loop, func, *args, **kwargs)

        else:
            worker = yield from self._tqueue.get()
            future, waiter = worker.submit(func, *args, **kwargs)
            waiter = asyncio.wrap_future(waiter, loop=self._loop)
            waiter.add_done_callback(lambda f: self._tqueue.put_nowait(worker))

        return (yield from asyncio.wrap_future(future, loop=self._loop))

    def schedule(self, interval, func, *args, **kwargs):
        """Run given func/coro periodically."""
        if isinstance(interval, datetime.timedelta):
            interval = interval.total_seconds()

        if not isinstance(interval, float):
            interval = float(interval)

        async def scheduler():
            while self.is_running():
                await self.submit(func, *args, **kwargs)
                await asyncio.sleep(interval)

        self._schedules.append(asyncio.ensure_future(scheduler(), loop=self._loop))

    def __str__(self):
        return 'Donald [%s]' % self.params['num_threads']
